@@ -1,13 +1,13 @@
+# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
 import random
-from typing import Dict, Tuple, Callable, Any, Sequence, Union, Mapping, Optional
+from typing import Dict, Any, Sequence, Union, Optional
 from collections import OrderedDict
 
 import torch
 
-from e3nn import o3
+from e3nn.o3._irreps import Irreps
 
 from nequip.data import AtomicDataDict
-from nequip.utils import instantiate
 
 
 class GraphModuleMixin:
@@ -40,13 +40,15 @@ class GraphModuleMixin:
         # Coerce
         irreps_in = {} if irreps_in is None else irreps_in
         irreps_in = AtomicDataDict._fix_irreps_dict(irreps_in)
+
         # positions are *always* 1o, and always present
         if AtomicDataDict.POSITIONS_KEY in irreps_in:
-            if irreps_in[AtomicDataDict.POSITIONS_KEY] != o3.Irreps("1x1o"):
+            if irreps_in[AtomicDataDict.POSITIONS_KEY] != Irreps("1x1o"):
                 raise ValueError(
                     f"Positions must have irreps 1o, got instead `{irreps_in[AtomicDataDict.POSITIONS_KEY]}`"
                 )
-        irreps_in[AtomicDataDict.POSITIONS_KEY] = o3.Irreps("1o")
+        irreps_in[AtomicDataDict.POSITIONS_KEY] = Irreps("1o")
+
         # edges are also always present
         if AtomicDataDict.EDGE_INDEX_KEY in irreps_in:
             if irreps_in[AtomicDataDict.EDGE_INDEX_KEY] is not None:
@@ -55,6 +57,14 @@ class GraphModuleMixin:
                 )
         irreps_in[AtomicDataDict.EDGE_INDEX_KEY] = None
 
+        # atom types are also always present
+        if AtomicDataDict.ATOM_TYPE_KEY in irreps_in:
+            if irreps_in[AtomicDataDict.ATOM_TYPE_KEY] is not None:
+                raise ValueError(
+                    f"atom types must have irreps None, got instead `{irreps_in[AtomicDataDict.ATOM_TYPE_KEY]}`"
+                )
+        irreps_in[AtomicDataDict.ATOM_TYPE_KEY] = None
+
         my_irreps_in = AtomicDataDict._fix_irreps_dict(my_irreps_in)
 
         irreps_out = AtomicDataDict._fix_irreps_dict(irreps_out)
@@ -97,6 +107,19 @@ class GraphModuleMixin:
         self.irreps_in.update(irreps_in)
         self.irreps_out.update(irreps_out)
 
+    @torch.jit.unused
+    def _get_metadata_contributions(self) -> Dict[str, str]:
+        """Override to provide dynamic metadata at compilation time.
+
+        Modules can override this to contribute metadata based on their current state (e.g., learned parameters).
+        Called by GraphModel during compilation.
+        All values must be strings.
+
+        Returns:
+            Dict[str, str]: Metadata key-value pairs. Can override static config values (e.g., per_edge_type_cutoff) or add new keys.
+        """
+        return {}
+
     def _make_tracing_inputs(self, n):
         # We impliment this to be able to trace graph modules
         out = []
@@ -136,9 +159,9 @@ class SequentialGraphNetwork(GraphModuleMixin, torch.nn.Sequential):
             module_list = list(modules)
         # check in/out irreps compatible
         for m1, m2 in zip(module_list, module_list[1:]):
-            assert AtomicDataDict._irreps_compatible(
-                m1.irreps_out, m2.irreps_in
-            ), f"Incompatible irreps_out from {type(m1).__name__} for input to {type(m2).__name__}: {m1.irreps_out} -> {m2.irreps_in}"
+            assert AtomicDataDict._irreps_compatible(m1.irreps_out, m2.irreps_in), (
+                f"Incompatible irreps_out from {type(m1).__name__} for input to {type(m2).__name__}: {m1.irreps_out} -> {m2.irreps_in}"
+            )
         self._init_irreps(
             irreps_in=module_list[0].irreps_in,
             my_irreps_in=module_list[0].irreps_in,
@@ -151,72 +174,7 @@ class SequentialGraphNetwork(GraphModuleMixin, torch.nn.Sequential):
             modules = OrderedDict((f"module{i}", m) for i, m in enumerate(module_list))
         super().__init__(modules)
 
-    @classmethod
-    def from_parameters(
-        cls,
-        shared_params: Mapping,
-        layers: Dict[str, Union[Callable, Tuple[Callable, Dict[str, Any]]]],
-        irreps_in: Optional[dict] = None,
-    ):
-        r"""Construct a ``SequentialGraphModule`` of modules built from a shared set of parameters.
-
-        For some layer, a parameter with name ``param`` will be taken, in order of priority, from:
-          1. The specific value in the parameter dictionary for that layer, if provided
-          2. ``name_param`` in ``shared_params`` where ``name`` is the name of the layer
-          3. ``param`` in ``shared_params``
-
-        Args:
-            shared_params (dict-like): shared parameters from which to pull when instantiating the module
-            layers (dict): dictionary mapping unique names of layers to either:
-                  1. A callable (such as a class or function) that can be used to ``instantiate`` a module for that layer
-                  2. A tuple of such a callable and a dictionary mapping parameter names to values. The given dictionary of parameters will override for this layer values found in ``shared_params``.
-                Options 1. and 2. can be mixed.
-            irreps_in (optional dict): ``irreps_in`` for the first module in the sequence.
-
-        Returns:
-            The constructed SequentialGraphNetwork.
-        """
-        # note that dictionary ordered gueranteed in >=3.7, so its fine to do an ordered sequential as a dict.
-        built_modules = []
-        for name, builder in layers.items():
-            if not isinstance(name, str):
-                raise ValueError(f"`'name'` must be a str; got `{name}`")
-            if isinstance(builder, tuple):
-                builder, params = builder
-            else:
-                params = {}
-            if not callable(builder):
-                raise TypeError(
-                    f"The builder has to be a class or a function. got {type(builder)}"
-                )
-
-            instance, _ = instantiate(
-                builder=builder,
-                prefix=name,
-                positional_args=(
-                    dict(
-                        irreps_in=(
-                            built_modules[-1].irreps_out
-                            if len(built_modules) > 0
-                            else irreps_in
-                        )
-                    )
-                ),
-                optional_args=params,
-                all_args=shared_params,
-            )
-
-            if not isinstance(instance, GraphModuleMixin):
-                raise TypeError(
-                    f"Builder `{builder}` for layer with name `{name}` did not return a GraphModuleMixin, instead got a {type(instance).__name__}"
-                )
-
-            built_modules.append(instance)
-
-        return cls(
-            OrderedDict(zip(layers.keys(), built_modules)),
-        )
-
+    @torch.jit.unused
     def append(self, name: str, module: GraphModuleMixin) -> None:
         r"""Append a module to the SequentialGraphNetwork.
 
@@ -229,31 +187,7 @@ class SequentialGraphNetwork(GraphModuleMixin, torch.nn.Sequential):
         self.irreps_out = dict(module.irreps_out)
         return
 
-    def append_from_parameters(
-        self,
-        shared_params: Mapping,
-        name: str,
-        builder: Callable,
-        params: Dict[str, Any] = {},
-    ) -> None:
-        r"""Build a module from parameters and append it.
-
-        Args:
-            shared_params (dict-like): shared parameters from which to pull when instantiating the module
-            name (str): the name for the module
-            builder (callable): a class or function to build a module
-            params (dict, optional): extra specific parameters for this module that take priority over those in ``shared_params``
-        """
-        instance, _ = instantiate(
-            builder=builder,
-            prefix=name,
-            positional_args=(dict(irreps_in=self[-1].irreps_out)),
-            optional_args=params,
-            all_args=shared_params,
-        )
-        self.append(name, instance)
-        return
-
+    @torch.jit.unused
     def insert(
         self,
         name: str,
@@ -311,44 +245,6 @@ class SequentialGraphNetwork(GraphModuleMixin, torch.nn.Sequential):
 
         return
 
-    def insert_from_parameters(
-        self,
-        shared_params: Mapping,
-        name: str,
-        builder: Callable,
-        params: Dict[str, Any] = {},
-        after: Optional[str] = None,
-        before: Optional[str] = None,
-    ) -> None:
-        r"""Build a module from parameters and insert it after ``after``.
-
-        Args:
-            shared_params (dict-like): shared parameters from which to pull when instantiating the module
-            name (str): the name for the module
-            builder (callable): a class or function to build a module
-            params (dict, optional): extra specific parameters for this module that take priority over those in ``shared_params``
-            after: the name of the module to insert after
-            before: the name of the module to insert before
-        """
-        if (before is None) is (after is None):
-            raise ValueError("Only one of before or after argument needs to be defined")
-        elif before is None:
-            insert_location = after
-        else:
-            insert_location = before
-        idx = list(self._modules.keys()).index(insert_location) - 1
-        if before is None:
-            idx += 1
-        instance, _ = instantiate(
-            builder=builder,
-            prefix=name,
-            positional_args=(dict(irreps_in=self[idx].irreps_out)),
-            optional_args=params,
-            all_args=shared_params,
-        )
-        self.insert(after=after, before=before, name=name, module=instance)
-        return
-
     # Copied from https://pytorch.org/docs/stable/_modules/torch/nn/modules/container.html#Sequential
     # with type annotations added
     def forward(self, input: AtomicDataDict.Type) -> AtomicDataDict.Type:
